# CPU Examples {#cpu_dev}

## Namespace and header files
The monolish namespace contains monolish::vector and monolish::view1D.
monolish has the following 7 namespaces in addition to the monolish namespace:
- monolish::matrix : Provides classes for Dense and Sparse matrices. This can be used by including other header files.
- monolish::blas : Provides a monolithic BLAS API that eliminates dependencies on data types, matrix format, and hardware-specific APIs. This is included in `monolish_blas.hpp`.
- monolish::vml : Provides the calculation of mathematical functions to each element of a vector. This is an open-source implementation of the [VML functions included in Intel MKL](https://software.intel.com/content/www/us/en/develop/articles/new-mkl-vml-api.html). This is included in `monolish_vml.hpp`.
- monolish::equation : Provides a solution for linear equations.	This is included in `monolish_equation.hpp`.
- monolish::standard_eigen : Provides solutions to standard eigenvalue problems. This is included in `monolish_eigen.hpp`.
- monolish::generalized_eigen : Provides a solution to the generalized eigenvalue problem. This is included in `monolish_generalized_eigen.hpp`.
- monolish::util : Provides utility functions. This can be used by including other header files.

This chapter describes a sample program using monolish that runs on the CPU.

Sample programs that run on the GPU can be found at `examples/`. They work on CPU and GPU.
The programs running on the GPU are described in the next chapter.

Sample programs for CPU can be found at `examples/cpu_only`. They only work on the CPU.

## Compute innerproduct 
First, a simple inner product program is shown below:

\include cpu_only/innerproduct/innerproduct.cpp

This program can be compiled by the following command.
```
g++ -O3 innerproduct.cpp -o innerproduct_cpu.out -lmonolish_cpu
```

The following command runs this.
``` 
./innerproduct_cpu.out
```

A description of this program is given below:
- A monolish::vector can be declared like a std::vector.
- As an extension, monolish::vector can also create random vectors.
- The inner product function is monolish::blas::dot(). It does not need type-dependent function names, for example `sdot` or `ddot`.
- For the BLAS library called inside monolish::blas::dot(), see [here](@ref oplist).
- At the end of the function, the memory of monolish::vector is automatically released by the destructor.

This program is executed in parallel. 
The number of threads is specified by the `OMP_NUM_THREADS` environment variable.

## Solve Ax=b
The following is a sample program that solves a linear equation; Ax=b using the conjugate gradient method with jacobi preconditioner.

\include cpu_only/cg/cg.cpp

### Matrix creation

monolish::matrix::COO has `editable` attribute, so users can edit any element using @ref monolish::matrix::COO.insert() "insert()" and monolish::matrix::COO.at() "at()" functions. COO can be created by giving array pointer or file name to the constructor, too.
This sample program creates a matrix from a file.
The input file format is [MatrixMarket format](https://math.nist.gov/MatrixMarket/formats.html).
MatrixMarket format is a common data format for sparse matrices. [scipy](https://docs.scipy.org/doc/scipy/reference/generated/scipy.io.mmwrite.html#scipy.io.mmwrite) can also output matrices in MatrixMarket format.

monolish::matrix::CRS can easily be generated by taking COO as an argument to the constructor. 

### Solving Ax=b
monolish proposes a new way to implement linear solvers.

Applying preconditioning means combining multiple solvers. In other words, the solver and preconditioner in the Krylov subspace method are essentially the same thing.

The solver classes included in monolish::equation all have the following functions:
- @ref monolish::equation::CG.solve() "solve()" : Solve the simultaneous linear equations.
- @ref monolish::equation::CG.create_precond() "create_precond()" : Create the preconditioning matrix M.
- @ref monolish::equation::CG.apply_precond() "apply_precond()" : Apply preconditioner according to the created preprocessing matrix M.
- @ref monolish::equation::CG.set_create_precond() "set_create_precond()" : Register another class' create_precond to the class as a function to create the preprocessing matrix.
- @ref monolish::equation::CG.set_apply_precond() "set_apply_precond()" : Register "apply_precond" of another class to the class as a function to apply preprocessing.

The class that executes `solve()` execute the registered `create_precond()` and `apply_precond()`.
If no preconditioner is registered, it calls monolish::equation::none as "none preconditioner".

By being able to register the creation and application of preconditioners separately, users can use the preconditioner matrix in different ways.

In the current version, the create_precond() and apply_precond() functions of some classes do not work.
In the future, we will make these functions work in all classes.
This implementation would be very efficient for multi-grid methods.

See [here](@ref solverlist) for a list of solvers.

## Improve solver program
The program in the previous chapter has the matrix storage format, data type, solver name, and preconditioner name explicitly specified all over the program.

If users want to change them depending on the input matrix, this implementation requires a lot of program changes.

Since monolish is designed so that the matrix and solver classes all have the same interface, these changes can be eliminated by templating.

A templated program is shown below.

\include cpu_only/templated_solver/solver.cpp

This program is templated so that the function that solves Ax=b is not oblivious to the matrix storage format or data type.
